/* -*- Mode: C; c-basic-offset:4 ; indent-tabs-mode:nil ; -*- */
/*
 *  (C) 2001 by Argonne National Laboratory.
 */

#include "dmem.h"
#include <stdarg.h>
#include <string.h>
#include <errno.h>

/* Allow tracing or not */
/* Set DMEM_TRACE to 0 to compile out all tracing */
#define DMEM_TRACE 1

/* Initialize specifiic tracing flags in DMEM_Init_dbg_flags, near end of this file. */

#if DMEM_TRACE == 1
#define DMEM_p(flag, ...) DMEM_p_fcn(flag, __VA_ARGS__)
#else
#define DMEM_p(flag, ...) {}
#endif

/* tags for service requests */
#define DMEM_GET             101   /* app to server */
#define DMEM_PUT             102   /* app to server */
#define DMEM_FREE            103   /* app to server */
#define DMEM_UPDATE          104   /* app to server */
#define DMEM_COPY            116   /* app to server */
#define DMEM_GET_PART        117   /* app to server */
/* other message tags */
#define DMEM_SHUTDOWN        105
#define DMEM_PUT_WHERE_REQ   106   /* app to manager */
#define DMEM_PUT_WHERE_RPLY  107   /* manager to app */
#define DMEM_PUT_ADDR        108   /* server to app  */
#define DMEM_INIT_SIZE       109
#define DMEM_DUMP_GLOBMEMTAB 111
#define DMEM_UPDATE_DATA     112
#define DMEM_GET_PART_DATA   118   /* server to app */
#define DMEM_UPDATE_RPLY     113   /* obsolete */
#define DMEM_GET_RPLY        114   /* server to app  */
#define DMEM_COPY_RPLY       115   /* server to app  */
#define DMEM_COPY_WHERE_REQ  128   /* server to manager */
#define DMEM_COPY_WHERE_RPLY 129   /* manager to server */
#define DMEM_ACK             119   /* manager to app */
#define DMEM_MEM_QUERY       120
#define DMEM_MEM_QUERY_REPLY 121   /* manager to app */
#define DMEM_MEM_FREE        122
#define DMEM_STATS           123
#define DMEM_FREE_CONF       124   /* server to app */
#define DMEM_MEM_UPDATE      125   /* app or server to manager */
#define DMEM_MUR             126   /* mem update reply */
#define DMEM_DECR_LOCMEM     127   /* app to its own server */
#define DMEM_COPY_DATA       130   /* server to server */
#define DMEM_COPY_DATA_RPLY  131   /* server to server */
#define DMEM_SPACE_QUERY_RPLY 133  /* server to server */
#define DMEM_PUT_DATA        134   /* app to server    */
#define DMEM_COPY_QUERY      135   /* server to server */
#define DMEM_MIN_LWM         136   /* server to manager */
#define DMEM_GET_RPLY_ERROR  137   /* server to app */
#define DMEM_MEM_UPDATE2     138   /* app to manager */
#define DMEM_MUR2            139   /* manager to app */
#define DMEM_STORE_ADDR      140   /* app to server */
#define DMEM_STORE_ADDR_RPLY 141   /* server to app */

#define SRVR 0
#define ADDR 1
#define SIZE 2

#define ORIGINAL 5
#define NOLOCAL  6

/* Debug print flags.  These are set in DMEM_Init_dbg_flags, near the end of this file */
static int Get_dbg_flag;
static int Put_dbg_flag;
static int Put_remote_dbg_flag;
static int Copy_dbg_flag;
static int Copy_remote_dbg_flag;
static int Free_dbg_flag;
static int Update_dbg_flag;
static int Get_part_dbg_flag;
static int Server_dbg_flag;
static int Manager_dbg_flag;
static int Memory_dbg_flag;
static int Error_return_flag;
static int Update_counter_init;
static int mgr_update_counter;
static int Selection_algorithm;
/* The global memory table maintains the size of memory available on each server, and also
   the number of times this is updated, for each server */ 
typedef struct {
    MPI_Aint size;                /* size available at any one time on a server */
    MPI_Aint num_updates;         /* number of times size has been updated */
    MPI_Aint initial_size;
} globmemtabent;
globmemtabent *globmemtab;        /* should only be accessed by manager */
static MPI_Aint globmemtotal;     /* total memory available to DMEM at any time*/
static MPI_Aint globmemtotal_lwm; /* low water mark of such memory */
static MPI_Aint globmemtotal_hwm; /* high water mark of such memory  ( = init total size) */

/* The local memory table has the addresses and sizes of all memory segments allocated on a server */
typedef struct {
    MPI_Aint address;
    MPI_Aint size;
    int      usage_flag;        /* -1 = initial value, never used.  -2 = used but freed.  -3 = in use */
} locmemtabent;

void *DMEM_Manager(void *);
void *DMEM_Server(void *);

void DMEM_Store_address(MPI_Aint, MPI_Aint);
int  DMEM_checkcopy(MPI_Aint, MPI_Aint, MPI_Aint);
int  DMEM_allocchk(MPI_Aint, MPI_Aint, MPI_Aint, MPI_Aint);
int  DMEM_checkoverlap(MPI_Aint, MPI_Aint, int);
void DMEM_Dump_gmem_table(globmemtabent *);
void DMEM_Dump_gmem_table_stats(globmemtabent *);
int  DMEM_Dumplocmemtab();
int  DMEM_Expandlocmemtab();
int  DMEM_Dumpglobmemtab();
char* DMEM_Tag_string(int);
int DMEM_Find_max(globmemtabent *, int);
int DMEM_Lookup(MPI_Aint);
void DMEM_Init_dbg_flags();
void *DMEM_amalloc(size_t);
char* a2s(long);
char* n2s(long);

#define LOCMEMTABSIZE_INIT 200
static int locmemtabsize;       /* initialized to LOCMEMTABSIZE_INIT, can change if locmemtab is expanded */
locmemtabent *locmemtab;
/*
dyamically allocated now
static locmemtabent locmemtab[locmemtabsize];
*/
static int locmemtabsize_hwm;   /* note locmemTABsize, not locmemsize */
static int max_locmemtabsize;
static MPI_Comm dmem_comm, dapp_comm;;
static int dmem_rank, dmem_size;
static int dmem_dbgprintf_flag;
static MPI_Aint locmemsize;     /* amount of memory currently available */
static MPI_Aint locmemsize_lwm; /* dynamically updated */
static MPI_Aint locmemsize_hwm; /* set to init size passed in*/
static MPI_Datatype handle_type;
static int num_irecvs;
static int dmem_mgr;            /* which process is manager (case not 0 not tested) */
static char unknown_tag[20];    /* place to return unrecognized tag value */
static int num_fat_servers;     /* number of servers given positive amounts of memory */
static pthread_t master_thread_id; /* thread that calls pthread_create */
static pthread_t thread_id;     /* thread created for DMEM, either manager or server */

/* statistics maintained by clients locally */
static int num_puts         = 0;
static int num_local_puts   = 0;
static int num_gets         = 0;
static int num_frees        = 0;
static int num_updates      = 0;
static int num_get_parts    = 0;
static int num_copies       = 0;
static int num_local_copies = 0;

static MPI_Aint bytes_put          = 0;
static MPI_Aint bytes_local_put    = 0;
static MPI_Aint bytes_copied       = 0;
static MPI_Aint bytes_local_copied = 0;
static MPI_Aint bytes_updated      = 0;
static MPI_Aint bytes_got          = 0;
static MPI_Aint bytes_local_got    = 0;
static MPI_Aint bytes_got_part     = 0;
static MPI_Aint bytes_freed        = 0;

/* accumulated in finalize */
static int total_puts, total_local_puts, total_gets, total_frees;
static int total_updates, total_get_parts, total_copies, total_local_copies;
static MPI_Aint total_bytes_put, total_bytes_local_put, total_bytes_copied;
static MPI_Aint total_bytes_local_copied, total_bytes_updated, total_bytes_got;
static MPI_Aint total_bytes_local_got, total_bytes_got_part, total_bytes_freed;
static MPI_Aint total_locmemsize, total_locmemsize_hwm, total_locmemsize_lwm, min_locmemsize_lwm;
static int      total_locmemtabsize, max_locmemtabsize;
static MPI_Aint num_mem_calls = 0;
static MPI_Aint total_used    = 0;
static MPI_Aint avg_used;

int DMEM_Init(MPI_Comm user_comm, MPI_Aint init_memsize, int num_irecvs_to_post)
{
    int i, rc, thread_level;
    MPI_Status status;

    MPI_Comm_dup(user_comm, &dmem_comm);
    MPI_Comm_rank(dmem_comm, &dmem_rank);
    MPI_Comm_size(dmem_comm, &dmem_size);
    MPI_Comm_dup(dmem_comm, &dapp_comm); /* used for communication with application thread */

    MPI_Type_vector(3, 1, 1, MPI_AINT, &handle_type);
    MPI_Type_commit(&handle_type);

    dmem_dbgprintf_flag = 1;                         /* debugging on by default */
    DMEM_Init_dbg_flags();                           /* initialize individual flags */
    dmem_mgr            = 0;                         /* set which process is manager */
    num_irecvs          = num_irecvs_to_post;        /* to pass to manager thread */
    locmemsize          = init_memsize;              /* local copy of memsize for non-managers */
    locmemsize_lwm      = locmemsize;                /* low-water mark maintained locally */
    locmemsize_hwm      = init_memsize;              /* high_water mark set once at init time */
    locmemtabsize       = LOCMEMTABSIZE_INIT;        /* initial size of locmemtab */
    locmemtab           = DMEM_amalloc(locmemtabsize * sizeof(locmemtabent));
    if (dmem_rank == dmem_mgr)
        locmemsize = locmemsize_hwm = locmemsize_lwm = 0; /* make sure manager is never treated as a server */

    MPI_Query_thread(&thread_level);
    if (thread_level != MPI_THREAD_MULTIPLE) {
        printf("thread level must be MPI_THREAD_MULTIPLE for DMEM\n");
        DMEM_Abort(135);
    }

    /* Initialize locmemtab. Manager doesn't use locmemtab, but we initialize it anyway,
       to make it simpler to dump all the locmemtabs. */
    for (i = 0; i < locmemtabsize; i++) {
        locmemtab[i].address    = -1;
        locmemtab[i].size       = -1;
        locmemtab[i].usage_flag = -1;
    }
    locmemtabsize_hwm = 0;      /* index beyond last entry */

    if (dmem_rank != dmem_mgr) {
        /* Send initial memory size to manager */
        MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_INIT_SIZE, dmem_comm);
    } else {                    /* manager */
        /* Initialize global memory data structures */
        globmemtab = DMEM_amalloc(dmem_size * sizeof(globmemtabent));
        globmemtotal = 0;
        num_fat_servers = 0;
        for (i = 0; i < dmem_size; i++) {
            if (i != dmem_mgr) {
                MPI_Recv(&globmemtab[i].size, 1, MPI_AINT, i, DMEM_INIT_SIZE, dmem_comm, &status);
                DMEM_p(Manager_dbg_flag, "manager received init size %ld from server rank %d \n", globmemtab[i].size, i);
                globmemtab[i].initial_size = globmemtab[i].size;
                globmemtab[i].num_updates = 0;
                globmemtotal += globmemtab[i].size;
                if (globmemtab[i].size > 0)
                    num_fat_servers++; /* accumulate number of servers with positive memory, for computing averages */
            }
        }
    }
    /* Make sure all data structures are initialized before starting processing */
    MPI_Barrier(dmem_comm);

    master_thread_id = pthread_self();

    if (dmem_rank == dmem_mgr) { 
        rc = pthread_create(&thread_id, NULL, DMEM_Manager, NULL);
        if (rc != 0) {
            printf("could not create manager thread; pthread_create failed\n");
            DMEM_Abort(619);
        }
    }
    else {                       
        rc = pthread_create(&thread_id, NULL, DMEM_Server, NULL);
        if (rc != 0) {
            printf("could not create server thread; pthread_create failed\n");
            DMEM_Abort(620);
        }
    }
    MPI_Barrier(dmem_comm);
    return 0;
}

int DMEM_Finalize()
{
    MPI_Aint dummy;         /* message content is just tag */
    MPI_Comm positive_comm; /* communicator of processes passed non-zero memory */
    MPI_Status status;
    int color, key;         /* for creating communicator of processes passed non-0 memory at init */
    int myposrank;          /* my rank in positive_comm */

    /* gather statistics */  /* Possible optimization: put these in a vector and do one reduce */
    MPI_Reduce(&num_puts, &total_puts, 1, MPI_INT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&num_local_puts, &total_local_puts, 1, MPI_INT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&num_gets, &total_gets, 1, MPI_INT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&num_frees, &total_frees, 1, MPI_INT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&num_updates, &total_updates, 1, MPI_INT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&num_get_parts, &total_get_parts, 1, MPI_INT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&num_copies, &total_copies, 1, MPI_INT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&num_local_copies, &total_local_copies, 1, MPI_INT, MPI_SUM, dmem_mgr, dmem_comm);

    MPI_Reduce(&bytes_put, &total_bytes_put, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&bytes_local_put, &total_bytes_local_put, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&bytes_copied, &total_bytes_copied, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&bytes_local_copied, &total_bytes_local_copied, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&bytes_got, &total_bytes_got, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&bytes_local_got, &total_bytes_local_got, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&bytes_got_part, &total_bytes_got_part, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&bytes_updated, &total_bytes_updated, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&bytes_freed, &total_bytes_freed, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);

    MPI_Reduce(&locmemsize_hwm,    &total_locmemsize_hwm, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&locmemtabsize_hwm, &total_locmemtabsize,  1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&locmemtabsize_hwm, &max_locmemtabsize,    1, MPI_INT,  MPI_MAX, dmem_mgr, dmem_comm);
    MPI_Reduce(&locmemsize_lwm,    &total_locmemsize_lwm, 1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    MPI_Reduce(&locmemsize,        &total_locmemsize,     1, MPI_AINT, MPI_SUM, dmem_mgr, dmem_comm);
    /* Create communicator of processes that are initialized with non-zero memory */
    if (locmemsize_hwm == 0)    /* memory passed to this process at init was 0, or this is manager */
        color = key = 0;
    else
        color = key = 1;
    MPI_Comm_split(dmem_comm, color, key, &positive_comm);
    if (color == 1) {
        MPI_Reduce(&locmemsize_lwm, &min_locmemsize_lwm, 1, MPI_AINT, MPI_MIN, 0, positive_comm); 
        MPI_Comm_rank(positive_comm, &myposrank);
        if (myposrank == 0)
            MPI_Send(&min_locmemsize_lwm, 1, MPI_AINT, dmem_mgr, DMEM_MIN_LWM, dmem_comm);
    }

    if (dmem_rank == dmem_mgr) {
        printf("Puts: %s; ", a2s(total_puts));
        printf("Local puts: %s; ", a2s(total_local_puts));
        printf("Frees: %s; ", a2s(total_frees));
        printf("Gets: %s;\n", a2s(total_gets));
        printf("Updates: %s; ", a2s(total_updates));
        printf("Partial Gets: %s; ", a2s(total_get_parts));
        printf("Copies: %s; ", a2s(total_copies));
        printf("Local copies: %s\n", a2s(total_local_copies));

        printf("Bytes put: %s; ", n2s(total_bytes_put));
        printf("Bytes put locally: %s; ", n2s(total_bytes_local_put));
        printf("Bytes put non-locally: %s\n", n2s(total_bytes_put - total_bytes_local_put));
        printf("Bytes copied: %s; ", n2s(total_bytes_copied));
        printf("Bytes copied locally: %s; ", n2s(total_bytes_local_copied));
        printf("Bytes copied non-locally: %s\n", n2s(total_bytes_copied - total_bytes_local_copied));
        printf("Bytes gotten: %s; ", n2s(total_bytes_got));
        printf("Bytes locally gotten: %s; ", n2s(total_bytes_local_got));
        printf("Bytes partially gotten: %s\n", n2s(total_bytes_got_part));
        printf("Bytes updated: %s; ", n2s(total_bytes_updated));
        printf("Bytes freed: %s\n", n2s(total_bytes_freed));

        printf("Local memory size totals across servers: max (hwm, init) = %s, lwm (min) = %s, current = %s\n",
               n2s(total_locmemsize_hwm), n2s(total_locmemsize_lwm), n2s(total_locmemsize));
        printf("Local memory size averages across fat servers: max (hwm, init) = %s, lwm (min) = %s, current = %s\n",
               n2s(total_locmemsize_hwm/(num_fat_servers)),
               n2s(total_locmemsize_lwm/(num_fat_servers)),
               n2s(total_locmemsize/(num_fat_servers)));
        printf("Max local memory table size hwm across servers      = %d\n", max_locmemtabsize);
        printf("Average local memory table size hwm across servers  = %d\n", total_locmemtabsize/(dmem_size - 1));

        printf("Local memory minumum lwm across servers given memory: min_lwm = %s\n", n2s(min_locmemsize_lwm));
        printf("Dynamic global memory statistics from manager (approximate): hwm (init) = %s, lwm = %s, current = %s\n",
               n2s(globmemtotal_hwm), n2s(globmemtotal_lwm), n2s(globmemtotal));

        /* print summary statistics from globmemtab */
        DMEM_Dump_gmem_table_stats(globmemtab);
        /* print statistics on total mem calls, average memory used */
        /* printf("New: Global amount of DMEM memory available at end: = %s, lwm = %s\n",
               n2s(globmemtotal), n2s(globmemtotal_lwm));
        */

        /*
        printf("Memory usage (BOGUS): Calls: %s, ", a2s(num_mem_calls));
        printf("Total memory in use at end: %s, ", a2s(total_used));
        printf("Average memory in use: %s\n", a2s(avg_used));
        */
     }

    /* fprintf(stderr, "Local memory summary for rank %d : init (max) = %ld; lwm = %ld, currently available = %ld\n",
       dmem_rank, locmemsize_hwm, locmemsize_lwm, locmemsize); */

    MPI_Send(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_SHUTDOWN, dmem_comm);
    pthread_join(thread_id, NULL);
    return 0;
}                                       /* end DMEM_Finalize */

/* manager, mainly responsible for receiving and storing memory capacity from other
   processes, then keeping track of approximate available memory on each server, and
   telling app processes where to store packages */

void * DMEM_Manager(void * dummy_arg)
{
    int           i, j, k, shutdowns_received, done, source_received, tag_received;
    MPI_Status    status, *statusarray;
    MPI_Request   *requestarray;
    int           *indexarray;
    int           count, maxcount = 0;
    MPI_Aint      *bufferarray, data_received, dummy = 1;
    double        manager_start_time, manager_total_time = 0.0;
    double        wait_start_time, wait_total_time = 0.0;


    printf("Rank %d master thread = %d, manager_thread = %d\n",
           dmem_rank, (int) master_thread_id, (int) pthread_self());

    globmemtotal_lwm = globmemtotal_hwm = globmemtotal;
    DMEM_p(Memory_dbg_flag, "Total memory assigned to DMEM at Init = %ld\n", globmemtotal);
    globmemtab[dmem_mgr].size = 0;            /* manager contributes no memory */
    globmemtab[dmem_mgr].initial_size = 0;    
    globmemtab[dmem_mgr].num_updates = 0;     /* shouldn't get any updates */

    shutdowns_received = 0;
    /* Post a bunch of irecvs */
    requestarray = (MPI_Request *) DMEM_amalloc(num_irecvs * sizeof(MPI_Status));
    bufferarray  = (MPI_Aint *) DMEM_amalloc(num_irecvs * sizeof(MPI_Aint));
    statusarray  = (MPI_Status *) DMEM_amalloc(num_irecvs * sizeof(MPI_Status));
    indexarray   = (int *) DMEM_amalloc(num_irecvs * sizeof(int));
    for (i = 0; i < num_irecvs; i++) 
        MPI_Irecv(&bufferarray[i], 1, MPI_AINT, MPI_ANY_SOURCE, MPI_ANY_TAG, dmem_comm, &requestarray[i]);

    done = 0;
    manager_start_time = MPI_Wtime();
    while (!done) {                      /* main manager loop */
        /* Wait for some to complete */
        wait_start_time = MPI_Wtime();
        MPI_Waitsome(num_irecvs, requestarray, &count, indexarray, statusarray);
        DMEM_p(0, "Irecvs completed on Waitsome = %d\n", count);
        wait_total_time += (MPI_Wtime() - wait_start_time);
        /* process those that completed */
        if (count > maxcount) maxcount = count;
        for (i = 0; i < count; i++) {
            j = indexarray[i];
            /* copy out the key parts of the message */
            data_received   = bufferarray[j];
            tag_received    = statusarray[i].MPI_TAG;
            source_received = statusarray[i].MPI_SOURCE;

            DMEM_p(Manager_dbg_flag, "Manager received request %s from rank %d, data = %ld\n",
                   DMEM_Tag_string(tag_received), source_received, data_received);
            if (tag_received == DMEM_SHUTDOWN) {
                int cancel_flag;
                MPI_Status cancel_status;

                shutdowns_received++;
                /* once all such messages from have been received and sent,
                   send (dummy) shutdown messages to all servers and return */
                if (shutdowns_received == dmem_size) {
                    for (k = 0; k < dmem_size; k++) {
                        if (k != dmem_mgr) {
                            MPI_Send(&dummy, 1, MPI_AINT, k, DMEM_SHUTDOWN, dmem_comm);
                        }
                    }
                    /* skip this for the time being; should be unnecessary anyway
                    for (k = 0; k < num_irecvs; k++) {
                        MPI_Cancel(&requestarray[k]);
                        MPI_Wait(&requestarray[k], &cancel_status);
                        MPI_Test_cancelled(&cancel_status, &cancel_flag);
                    } */
                    done = 1;
                }
            }
            else if (tag_received == DMEM_MIN_LWM) {
                min_locmemsize_lwm = data_received;
            }
            else if (tag_received == DMEM_DUMP_GLOBMEMTAB) {
                DMEM_Dump_gmem_table(globmemtab);
                MPI_Send(&dummy, 1, MPI_AINT, source_received, DMEM_ACK, dapp_comm); /* ack */
            }
            else if (tag_received == DMEM_STATS) {
                DMEM_Dump_gmem_table_stats(globmemtab);
                MPI_Send(&dummy, 1, MPI_AINT, source_received, DMEM_ACK, dapp_comm); /* ack */
            }
            else if (tag_received == DMEM_PUT_WHERE_REQ) {
                MPI_Aint size;
                int maxloc;

                size = data_received;

                /* Manager currently just selects server with most memory available. */
                maxloc = DMEM_Find_max(globmemtab, source_received); /* new use of exclude */
                if (maxloc == -1) {
                    fprintf(stderr, "DMEM: all memory exhausted, request size = %s\n", n2s(size));
                    DMEM_Abort(349);
                }
                if (globmemtab[maxloc].size - size < 0) {
                    printf("DMEM: Not enough memory at server with largest available memory\n");
                    printf("DMEM:    server rank %d, avail = %s, size needed = %s\n",
                           maxloc, n2s(globmemtab[maxloc].size), n2s(size));
                    DMEM_Abort(350);
                }
                DMEM_p(Put_dbg_flag, "manager received put-where request from rank %d, size=%ld, suggests %d\n",
                       source_received, size, maxloc);
                MPI_Send(&maxloc, 1, MPI_INT, source_received, DMEM_PUT_WHERE_RPLY, dapp_comm);
                DMEM_p(Put_dbg_flag, "manager handled put-where request from server rank %d, request size = %ld, newmemsize = %ld\n",
                       maxloc, size, globmemtab[maxloc].size);
            }
            else if (tag_received == DMEM_COPY_WHERE_REQ) { /* nearly identical to put-where case above*/
                MPI_Aint size;
                int maxloc;

                size = data_received;

                /* Manager currently just selects server with most memory available. */
                maxloc = DMEM_Find_max(globmemtab, source_received); /* new use of exclude */
                if (maxloc == -1) {
                    fprintf(stderr, "DMEM: all memory exhausted, request size = %s\n", n2s(size));
                    DMEM_Abort(351);
                }
                if (globmemtab[maxloc].size - size < 0) {
                    printf("Not enough memory at server with largest available memory\n");
                    printf("    server rank %d, avail = %s, size needed = %s\n",
                           maxloc, n2s(globmemtab[maxloc].size), n2s(size));
                    DMEM_Abort(352);
                }
                DMEM_p(Copy_dbg_flag, "manager received copy-where request from rank %d, size=%ld, suggests %d\n",
                       source_received, size, maxloc);
                DMEM_p(Copy_dbg_flag, "manager sending COPY_WHERE_RPLY %d to requester %d\n", maxloc, source_received);
                MPI_Send(&maxloc, 1, MPI_INT, source_received, DMEM_COPY_WHERE_RPLY, dmem_comm);
                DMEM_p(Copy_dbg_flag, "manager handled copy-where request from server rank %d, request size = %ld, newmemsize = %ld\n",
                       maxloc, size, globmemtab[maxloc].size);
            }
            else if (tag_received == DMEM_MEM_QUERY) {
                MPI_Aint request_size, current_size;
                int answer;

                DMEM_p(Memory_dbg_flag, "manager in MEM_QUERY from server rank %d, size = %ld\n",
                       source_received, data_received);
                request_size = data_received;
                /* look up memsize available to requester, answer whether it has enough memory
                   available for a local put or copy, update globmemtab for it if so */
                current_size = globmemtab[source_received].size;
                if (request_size <= current_size) {
                    /*printf("in manager mem query, subtracting %ld from %ld for process %d\n",
                      request_size, current_size, source_received); */
                    globmemtab[source_received].size -= request_size;
                    answer = 1;
                } else {
                    /* printf("in manager mem query, rejecting request from process %d\n", source_received); */
                    answer = 0;     /* requester will have to put or copy elsewhere rather than locally */
                    /* potential optimization: suggest alternate destination here instead of waiting for put */
                }
                num_mem_calls++;
                total_used += request_size;
                avg_used = total_used / num_mem_calls;
                MPI_Send(&answer, 1, MPI_INT, source_received, DMEM_MEM_QUERY_REPLY, dapp_comm);
            }
            else if (tag_received == DMEM_MEM_FREE) {
                MPI_Aint size_freed, current_size;
            
                printf("DMEM_FREE code in manager; should not be getting called since replaced by DMEM_MEM_UPDATE code\n");
                size_freed = data_received;
                current_size = globmemtab[source_received].size;
                globmemtab[source_received].size += size_freed;
                /* These next three lines are suspect */
                num_mem_calls++;
                total_used -= size_freed;;
                avg_used = total_used / num_mem_calls;
                DMEM_p(Free_dbg_flag, "manager in MEM_FREE, size freed = %ld, new size for server rank %d is %ld\n",
                       size_freed, source_received, globmemtab[source_received].size);
            }
            else if ((tag_received == DMEM_MEM_UPDATE) || (tag_received == DMEM_MEM_UPDATE2)) {
                MPI_Aint old_size;
                int dbg_flag;

                /* receive new locmemsize from either app or slave and update globmemtab, confirm */
                old_size = globmemtab[source_received].size;
                globmemtab[source_received].size = data_received;
                globmemtab[source_received].num_updates++;
                globmemtotal = globmemtotal + (data_received - old_size);
                if (globmemtotal < globmemtotal_lwm)
                    globmemtotal_lwm = globmemtotal;
                dbg_flag = Put_dbg_flag || Copy_dbg_flag || Free_dbg_flag || Memory_dbg_flag;
                DMEM_p(dbg_flag, "manager updated memsize for server rank %d from %ld to %ld\n", source_received,
                       old_size, globmemtab[source_received].size);
                if (tag_received == DMEM_MEM_UPDATE)
                    MPI_Send(&dummy, 1, MPI_AINT, source_received, DMEM_MUR, dmem_comm); /* confirm */
                else /* tag == DMEM_MEM_UPDATE2, so from app */
                    MPI_Send(&dummy, 1, MPI_AINT, source_received, DMEM_MUR2, dapp_comm); /* confirm */
            }
            else {
                fprintf(stderr, "DMEM: unexpected tag %s received in manager from rank %d \n",
                        DMEM_Tag_string(tag_received), source_received);
                DMEM_Abort(-523);
            }
            DMEM_p(Manager_dbg_flag, "Manager completed request %s from rank %d \n",
                   DMEM_Tag_string(tag_received), source_received);

            /* Repost the irecv */
            MPI_Irecv(&bufferarray[j], 1, MPI_AINT, MPI_ANY_SOURCE, MPI_ANY_TAG, dmem_comm, &requestarray[j]);
        }
    }
    manager_total_time = MPI_Wtime() - manager_start_time;
    printf("total manager time = %.3f, blocked in wait = %.3f, idle percentage = %.2f\n",
           manager_total_time, wait_total_time, (wait_total_time/manager_total_time) * 100);
    printf("      processing time = %.3f, active percentage = %.2f\n", manager_total_time - wait_total_time,
           ((manager_total_time - wait_total_time) / manager_total_time) * 100);
    /* printf("maximum irecvs complete at one time = %d\n", maxcount); */

    /* clean up */
    free(globmemtab);
    free(requestarray);
    free(bufferarray);
    free(statusarray);
    free(indexarray);

    /* printf("manager thread on rank %d returning\n", dmem_rank); */
    return NULL;                /* end manager */
}

void *DMEM_Server(void *arg)
{
    int i, done, count, tag;
    MPI_Status status;
    MPI_Aint dummy = 2;

    printf("Rank %d master thread = %d, server_thread = %d\n",
           dmem_rank, (int) master_thread_id, (int) pthread_self());

    /* main server loop */
    mgr_update_counter = Update_counter_init;
    done = 0;
    while (!done) {
        /* Receive a service request */
        MPI_Probe(MPI_ANY_SOURCE, MPI_ANY_TAG, dmem_comm, &status);
        tag = status.MPI_TAG;
        DMEM_p(Server_dbg_flag, "server received %s request\n", DMEM_Tag_string(tag));

        /* Split on tag for user operations */
        if        (status.MPI_TAG == DMEM_PUT) {
            /* This code handles remote puts.  Local puts are handled by the client code except for updating locmemtab.
               That is done by the server in response to a DMEM_STORE_ADDR message, handled below. */
            int answer;
            void *addr;
            MPI_Aint put_size, addr_aint;
            MPI_Status status1;

            DMEM_p(Put_dbg_flag, "server processing put from rank %d, locmemsize = %ld\n", status.MPI_SOURCE, locmemsize);

            MPI_Recv(&put_size, 1, MPI_AINT, status.MPI_SOURCE, DMEM_PUT, dmem_comm, &status1);
            /* Check whether we still have room;  manager thought so when it told the putter to do remote put here */
            if (put_size > locmemsize) {
                printf("Remote put from rank %d to server rank %d of size %ld failed; locmemsize = %ld; manager misdirected\n",
                       status.MPI_SOURCE, dmem_rank, put_size, locmemsize);
                /* Send current locmemsize to manager, who must be badly out of date */
                MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_MEM_UPDATE, dmem_comm);
                MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_MUR, dmem_comm, &status1);
                /* reject the remote put */
                answer = 0;     /* not enough room here */
                MPI_Send(&answer, 1, MPI_INT, status.MPI_SOURCE, DMEM_SPACE_QUERY_RPLY, dapp_comm);
            } else {
                answer = 1;
                MPI_Send(&answer, 1, MPI_INT, status.MPI_SOURCE, DMEM_SPACE_QUERY_RPLY, dapp_comm);
                addr = DMEM_amalloc((size_t) put_size);
                if (addr == NULL) {
                    printf("malloc of %ld bytes failed on server rank %d \n", put_size, dmem_rank);
                    DMEM_Abort(234);
                }
                /* Check whether new memory overlaps a chunk in locmemtab */
                /* DMEM_checkoverlap((MPI_Aint) addr, put_size, __LINE__); */

                locmemsize -= put_size; /* now guaranteed to remain positive */
                if (locmemsize < locmemsize_lwm) {
                    locmemsize_lwm = locmemsize;
                    if (locmemsize_lwm == 0)
                        DMEM_p(1, "locmemsize_lwm hit 0\n");
                }
                /* periodically update manager */
                if (mgr_update_counter == 0) {
                    /* Send new locmemsize to manager */
                    MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_MEM_UPDATE, dmem_comm);
                    MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_MUR, dmem_comm, &status1);
                    mgr_update_counter = Update_counter_init;
                } else {
                    mgr_update_counter--;
                }
                /* Receive the actual data */
                MPI_Recv(addr, put_size, MPI_BYTE, status.MPI_SOURCE, DMEM_PUT_DATA, dmem_comm, &status1);

                /* printf("server rank %d received size %d, %c,%c,...\n", dmem_rank, put_size,
                   ((char *) addr)[0], ((char *) addr)[1]); */

                /* store address and size in local mem table */
                DMEM_Store_address((MPI_Aint) addr, (MPI_Aint) put_size);
                addr_aint = (MPI_Aint) addr;
                MPI_Send(&addr_aint, 1, MPI_AINT, status.MPI_SOURCE, DMEM_PUT_ADDR, dapp_comm);
                DMEM_p(Put_dbg_flag, "server processed remote put from rank %d , addr = %ld, locmemsize = %ld\n",
                       status.MPI_SOURCE, addr_aint, locmemsize);
            }
        } else if (status.MPI_TAG == DMEM_STORE_ADDR) { /* server's role in a local put */
            MPI_Aint addrandsize[2];
            MPI_Status status1;
            int dummy;

            DMEM_p(Put_dbg_flag, "server receiving address from local put\n");
            MPI_Recv(addrandsize, 2, MPI_AINT, dmem_rank, DMEM_STORE_ADDR, dmem_comm, &status1);
            DMEM_Store_address(addrandsize[0], addrandsize[1]);
            MPI_Send(&dummy, 1, MPI_INT, dmem_rank, DMEM_STORE_ADDR_RPLY, dmem_comm);
            DMEM_p(Put_dbg_flag, "server replied to store_address message from local put\n");
        } else if (status.MPI_TAG == DMEM_COPY_QUERY) { /* very similar to DMEM_PUT, above */
            int answer;
            void *addr;
            MPI_Aint copy_size, addr_aint;
            MPI_Status status1;

            DMEM_p(Copy_dbg_flag, "server processing remote copy from rank %d, locmemsize = %ld\n",
                   status.MPI_SOURCE, locmemsize);
            /* receive message asking whether we really have space */
            MPI_Recv(&copy_size, 1, MPI_AINT, status.MPI_SOURCE, DMEM_COPY_QUERY, dmem_comm, &status1);
            DMEM_p(Copy_dbg_flag, "server received copy size %ld\n", copy_size);
            if (copy_size > locmemsize) {  /* detected misdirection */
                /* update manager */
                MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_MEM_UPDATE, dmem_comm);
                MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_MUR, dmem_comm, &status1);
                /* reject the remote copy */
                answer = 0;
                MPI_Send(&answer, 1, MPI_INT, status.MPI_SOURCE, DMEM_SPACE_QUERY_RPLY, dmem_comm);
            } else {
                answer = 1;
                MPI_Send(&answer, 1, MPI_INT, status.MPI_SOURCE, DMEM_SPACE_QUERY_RPLY, dmem_comm);
                DMEM_p(Copy_dbg_flag, "Sent SPACE_QUERY_RPLY, answer = 1, to rank %d\n", status.MPI_SOURCE);
                /* Prepare to receive data */
                addr = DMEM_amalloc((size_t) copy_size);
                if (addr == NULL) {
                    printf("malloc of %ld bytes failed in copy on server rank %d \n", copy_size, dmem_rank);
                    DMEM_Abort(235);
                }
                /* Check whether new memory overlaps a chunk in locmemtab */
                /* DMEM_checkoverlap((MPI_Aint) addr, copy_size, __LINE__); */

                locmemsize -= copy_size;
                if (locmemsize < locmemsize_lwm) {
                    locmemsize_lwm = locmemsize;
                    if (locmemsize_lwm == 0)
                        DMEM_p(1, "locmemize_lwm hit 0\n");
                }
                /* periodically update manager */
                if (mgr_update_counter == 0) {
                    /* Send new locmemsize to manager */
                    MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_MEM_UPDATE, dmem_comm);
                    MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_MUR, dmem_comm, &status1);
                    mgr_update_counter = Update_counter_init;
                } else {
                    mgr_update_counter--;
                }
                
                /* Receive data */
                DMEM_p(Copy_dbg_flag, "about to receive data from %d\n", status.MPI_SOURCE);
                MPI_Recv(addr, copy_size, MPI_BYTE, status.MPI_SOURCE, DMEM_COPY_DATA, dmem_comm, &status1);
                DMEM_p(Copy_dbg_flag, "received data from %d\n", status.MPI_SOURCE);

                /* printf("Server %d received size %d, %c,%c,...\n", dmem_rank, copy_size,
                   ((char *) addr)[0], ((char *) addr)[1]); */

                /* store address and size in local mem table */
                DMEM_Store_address((MPI_Aint) addr, (MPI_Aint) copy_size);
                addr_aint = (MPI_Aint) addr;
                MPI_Send(&addr_aint, 1, MPI_AINT, status.MPI_SOURCE, DMEM_COPY_DATA_RPLY, dmem_comm);
                DMEM_p(Copy_dbg_flag, "server processed copy from rank %d , addr = %ld, locmemsize = %ld\n",
                       status.MPI_SOURCE, addr_aint, locmemsize);
            }
        } else if (status.MPI_TAG == DMEM_UPDATE) {
            MPI_Aint handleoffsetlen[DMEM_HANDLE_SIZE+2];
            MPI_Status status1;
            int index;

            MPI_Recv(handleoffsetlen, 5, MPI_AINT,
                     status.MPI_SOURCE, DMEM_UPDATE, dmem_comm, &status1);
            DMEM_p(Update_dbg_flag, "server entering Update, addr = %ld\n", handleoffsetlen[ADDR]);
            index = DMEM_Lookup(handleoffsetlen[ADDR]);
            if (index == -1) {
                fprintf(stderr, "DMEM: Bad address in handle passed to DMEM_UPDATE by process %d\n", status.MPI_SOURCE);
                fprintf(stderr, "DMEM: handle: server rank=%d , addr=%ld, size=%ld\n",
                       dmem_rank, handleoffsetlen[ADDR], handleoffsetlen[SIZE]);
                DMEM_Abort(144);
            }
            else {
                MPI_Recv((char *) handleoffsetlen[ADDR] + handleoffsetlen[3], handleoffsetlen[4],
                         MPI_BYTE, status.MPI_SOURCE, DMEM_UPDATE_DATA, dmem_comm, &status1);
            }
            DMEM_p(Update_dbg_flag, "server exiting Update\n");
        } else if (status.MPI_TAG == DMEM_GET_PART) {
            MPI_Aint handleoffsetlen[DMEM_HANDLE_SIZE+2];
            MPI_Status status1;
            int index;

            MPI_Recv(handleoffsetlen, 5, MPI_AINT,
                     status.MPI_SOURCE, DMEM_GET_PART, dmem_comm, &status1);
            DMEM_p(Get_part_dbg_flag, "server entering Get_part, root addr = %ld, offset = %ld, len = %ld\n",
                   handleoffsetlen[ADDR], handleoffsetlen[3], handleoffsetlen[4]);

            index = DMEM_Lookup(handleoffsetlen[ADDR]);
            if (index == -1) {
                fprintf(stderr, "DMEM: Bad address in handle passed to DMEM_GET_PART by process %d\n", status.MPI_SOURCE);
                fprintf(stderr, "DMEM: handle: server rank=%d , addr=%ld, size=%ld\n",
                       dmem_rank, handleoffsetlen[ADDR], handleoffsetlen[SIZE]);
                DMEM_Abort(644);
            }
            else {
                MPI_Send((char *) handleoffsetlen[ADDR] + handleoffsetlen[3], handleoffsetlen[4],
                         MPI_BYTE, status.MPI_SOURCE, DMEM_GET_PART_DATA, dapp_comm);
            }
            DMEM_p(Get_part_dbg_flag, "server exiting Get_part\n");
        } else if (status.MPI_TAG == DMEM_FREE) {
            MPI_Aint addrandsize[2];
            MPI_Status status1;
            int i;

            MPI_Recv(addrandsize, 2, MPI_AINT, status.MPI_SOURCE, DMEM_FREE,
                     dmem_comm, &status1);
            /*pie       DMEM_p(Free_dbg_flag, "server entering FREE, addr = %ld, size = %ld, locmemsize = %ld\n",
              addrandsize[0], addrandsize[1], locmemsize); */
            /* clear entry in table */
            i = DMEM_Lookup(addrandsize[0]);
            if (i == -1) {
                fprintf(stderr, "DMEMd: Bad address (not in locmemtab %d) in handle passed to DMEM_Free by process %d\n",
                        dmem_rank, status.MPI_SOURCE);
                fprintf(stderr, "handle: server rank=%d , addr=%ld, size=%ld\n",
                       dmem_rank, addrandsize[0], addrandsize[1]);
                DMEM_Abort(143);
            }
            else {
                /*
                  locmemtab[i].address = -1;
                  locmemtab[i].size    = -1;
                */
                locmemtab[i].usage_flag = -2;
            }
            /* free the local memory */
            /*pie      DMEM_p(Free_dbg_flag, "server about to free %ld bytes on behalf of rank %d \n",
              addrandsize[1], status.MPI_SOURCE); */
            free((void *) addrandsize[0]);
            DMEM_p(Free_dbg_flag, "server successfully freed entry %d: %ld bytes at addr %ld\n", i,
                   addrandsize[1], addrandsize[0]);
            /* Update local memory size */
            locmemsize += addrandsize[1];
            /* Send new locmemsize to manager */
            MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_MEM_UPDATE, dmem_comm);
            MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_MUR, dmem_comm, &status1);
            /* Send free confirmation to client */
            MPI_Send(&dummy, 1, MPI_AINT, status.MPI_SOURCE, DMEM_FREE_CONF, dapp_comm);
            /*pie     DMEM_p(Free_dbg_flag, "server exiting FREE, locmemsize = %ld\n", locmemsize); */
        } else if (status.MPI_TAG == DMEM_GET) {
            /* receive message probed, find package in local mem tab, send to source */
            MPI_Aint addrandsize[2];
            MPI_Status status1; 
            int index;

            MPI_Recv(addrandsize, 2, MPI_AINT, status.MPI_SOURCE, DMEM_GET, dmem_comm, &status1);
            DMEM_p(Get_dbg_flag, "server entering DMEM_Get, addr = %ld, size = %ld\n", addrandsize[0], addrandsize[1]);
            index = DMEM_Lookup(addrandsize[0]);
            if (index == -1) {
                fprintf(stderr, "DMEM: Bad address %ld passed to DMEM_Get by rank %d, detected on server %d\n",
                       addrandsize[0], status.MPI_SOURCE, dmem_rank);
                DMEM_Dumplocmemtab(); /* to help with debugging */
                if (Error_return_flag == 1) {
                    MPI_Send((void *) NULL, 0, MPI_BYTE, status.MPI_SOURCE, DMEM_GET_RPLY_ERROR, dapp_comm);
                } else {
                    DMEM_Abort(237);
                }
            }
            else {
                MPI_Send((void *) addrandsize[0], addrandsize[1], MPI_BYTE,
                         status.MPI_SOURCE, DMEM_GET_RPLY, dapp_comm);
            }
            DMEM_p(Get_dbg_flag, "server exiting DMEM_Get\n");
        } else if (status.MPI_TAG == DMEM_COPY) {
            /* receive address, look up in local table, check if copy can be local */
            MPI_Status status1;
            MPI_Aint localaddr, copyaddr, addrandsrvr[2], size;
            int index, local_ok, dbg_flag;
            dmem_handle handle;

            MPI_Recv(&localaddr, 1, MPI_AINT, status.MPI_SOURCE, DMEM_COPY, dmem_comm, &status1);
            /*pie  DMEM_p(Copy_dbg_flag, "server entering copy, addr = %ld, locmemsize = %ld\n", localaddr, locmemsize); */
            index = DMEM_Lookup(localaddr);
            if (index == -1) {
                fprintf(stderr, "DMEM: Bad address passed to DMEM_Copy by process %d\n", status.MPI_SOURCE);
                DMEM_Abort(670);
            }
            size  = locmemtab[index].size;
            if (size <= locmemsize)
                local_ok = 1;
            else
                local_ok = 0;
            if (local_ok) {
                int rc;

            /*pie     DMEM_p(Copy_dbg_flag, "server making local copy since size = %ld, locmemsize = %ld\n",
              size, locmemsize); */
                /* make local copy, return new handle of copy */
                copyaddr = (MPI_Aint) DMEM_amalloc((size_t) size);
                if (copyaddr == (MPI_Aint) NULL) {
                    printf("malloc failed in local copy, size = %ld, locmemsize = %ld\n", size, locmemsize);
                    DMEM_Abort(147);
                }

                /* Check whether new memory overlaps a chunk in locmemtab */
                /* DMEM_checkoverlap(copyaddr, size, __LINE__); */

           /*pie      DMEM_p(Copy_dbg_flag, "server making local copy, copyaddr=%ld, localaddr=%ld, size=%ld\n",
             copyaddr, localaddr, size); */
                memcpy((void *) copyaddr, (void *) localaddr, (size_t) size);
                /* Check that copy was correct 
                rc = DMEM_checkcopy(localaddr, (MPI_Aint) copyaddr, size);
                if (!rc) {
                    DMEM_p(1, "checkcopy failed after memcpy in DMEM_Copy (local), rc = %d\n", rc);
                    DMEM_Abort(492);
                }
                */
                /* update local memory table */
                locmemsize -= size;
                if (locmemsize < locmemsize_lwm) {
                    locmemsize_lwm = locmemsize;
                    if (locmemsize_lwm == 0)
                        DMEM_p(1, "locmemsize_lwm hit 0\n");
                }

                /* Modify local memory table */
                DMEM_Store_address(copyaddr, size);

                /* periodically update manager */
                if (mgr_update_counter == 0) {
                    /* Send new locmemsize to manager */
                    MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_MEM_UPDATE, dmem_comm);
                    MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_MUR, dmem_comm, &status1);
                    mgr_update_counter = Update_counter_init;
                } else {
                    mgr_update_counter--;
                }

                /* Send new address (and this server) back to requester */
                addrandsrvr[0] = (MPI_Aint) copyaddr;
                addrandsrvr[1] = (MPI_Aint) dmem_rank;
                MPI_Send(addrandsrvr, 2, MPI_AINT, status.MPI_SOURCE, DMEM_COPY_RPLY, dapp_comm);
                DMEM_p(Copy_dbg_flag, "server made local copy, copyaddr=%ld, localaddr=%ld, size=%ld, new locmemsize=%ld\n",
                       copyaddr, localaddr, size, locmemsize);
            } else {            /* non-local copy necessary */
                MPI_Status status1;
                int srvr, dbg_flag, remote_answer, loop_count;
                MPI_Aint remote_address;

                dbg_flag = Copy_dbg_flag || Copy_remote_dbg_flag;
                DMEM_p(dbg_flag, "server doing remote copy since locmemsize = %ld, < pkg_len = %ld\n", locmemsize, size);
                /* Ask manager where to copy it, but first update manager's copy of this server's locmemsize */ 
                /* Possible optimization:  combine these two exchanges */

                remote_answer = 0;
                loop_count = 0;
                while (!remote_answer) {
                    MPI_Send(&size, 1, MPI_AINT, dmem_mgr, DMEM_COPY_WHERE_REQ, dmem_comm);
                    MPI_Recv(&srvr, 1, MPI_INT, dmem_mgr, DMEM_COPY_WHERE_RPLY, dmem_comm, &status1);
                    DMEM_p(dbg_flag, "server in DMEM_Copy, manager suggests %d as target for remote copy\n", srvr);
                    if (srvr == dmem_rank) { /* if manager suggests this server, we are out of space */
                        DMEM_p(1,"in remote copy, manager suggested copier (this process) as target - aborting\n");
                        DMEM_Abort(395);
                    } else {
                        /* Ask remote server if it really does have space */
                        MPI_Send(&size, 1, MPI_AINT, srvr, DMEM_COPY_QUERY, dmem_comm);
                        DMEM_p(Copy_dbg_flag, "about to receive notification from copy destination %d\n", srvr);
                        MPI_Recv(&remote_answer, 1, MPI_INT, srvr, DMEM_SPACE_QUERY_RPLY, dmem_comm, &status1);
                        DMEM_p(Copy_dbg_flag, "received notification from copy destination %d, answer = %d\n", srvr, remote_answer);
                        if (remote_answer == 1) {
                            /* Copy it:  send package to destination server and receive address back */
                            MPI_Send((void *) localaddr, size, MPI_BYTE, srvr, DMEM_COPY_DATA, dmem_comm);
                            MPI_Recv(&remote_address, 1, MPI_AINT, srvr, DMEM_COPY_DATA_RPLY, dmem_comm, &status1);
                            DMEM_p(dbg_flag, "server received copy address %ld from server rank %d rejections = %d\n",
                                   remote_address, srvr);
                            /* Now send handle info back to original copy requester */
                            addrandsrvr[0] = remote_address;
                            addrandsrvr[1] = (MPI_Aint) srvr;
                            DMEM_p(dbg_flag, "server sending addr and server back to orignal copy requester %d\n",
                                   status.MPI_SOURCE);
                            MPI_Send(addrandsrvr, 2, MPI_AINT, status.MPI_SOURCE, DMEM_COPY_RPLY, dapp_comm);
                        } else {
                            loop_count++;
                        }
                    }
                }
            }
        } else if (status.MPI_TAG == DMEM_SHUTDOWN) {
            /* receive shutdown message that was probed */
            MPI_Status status1;

            MPI_Recv(&dummy, 1, MPI_AINT, status.MPI_SOURCE, DMEM_SHUTDOWN, dmem_comm, &status1);
            done = 1;
        } else if (status.MPI_TAG == DMEM_DECR_LOCMEM) {
            MPI_Aint pkg_size;
            MPI_Status status1;
            
            MPI_Recv(&pkg_size, 1, MPI_AINT, status.MPI_SOURCE, DMEM_DECR_LOCMEM, dmem_comm, &status1);
            DMEM_p(Memory_dbg_flag, "in server DECR_LOCMEM, size = %ld\n", pkg_size);
            locmemsize -= pkg_size;     /* update local amount of local memory available */
            if (locmemsize < locmemsize_lwm) {
                locmemsize_lwm = locmemsize;
                if (locmemsize_lwm == 0)
                    DMEM_p(1, "locmemsize_lwm hit 0\n");
            }
            /* periodically update manager */
            if (mgr_update_counter == 0) {
                MPI_Status status1;
                /* Send new locmemsize to manager */
                MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_MEM_UPDATE, dmem_comm);
                MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_MUR, dmem_comm, &status1);
                mgr_update_counter = Update_counter_init;
            } else {
                mgr_update_counter--;
            }

        } else {
            fprintf(stderr, "DMEM server thread on rank %d probed unexpected DMEM operation %s from rank %d\n", dmem_rank,
                    DMEM_Tag_string(status.MPI_TAG), status.MPI_SOURCE);
            DMEM_Abort(-1000);
        }
        DMEM_p(Server_dbg_flag, "server completed %s request\n", DMEM_Tag_string(tag));
    }
    /* printf("server thread on rank %d returning\n", dmem_rank); */
    return NULL;                /* end server */
}

/* DMEM_Put puts the package defined by pkg_addr and pkg_len (in bytes) into dmem.
   The server rank, the local address there of the pkg, and the size are returned
   in the handle. */

int DMEM_Put(void *pkg_addr, MPI_Aint pkg_len, dmem_handle dh)
{
    /* Obtain from manager a destination to store it on, send to that
       server, and receive back the address where it stored it.  Then
       return server and address to caller */ /* FIX; it's more complicated now */

    MPI_Status status;
    MPI_Aint  remote_address, local_address, dummy; 
    MPI_Aint  addrandsize[2];
    int srvr, answer, dummy1, rc;
    

    /* Look to see locally if there is space to put it locally */
    if (pkg_len <= locmemsize)
        answer = 1;
    else
        answer = 0;
    if (answer == 1) {          /* enough room to do local put */
        DMEM_p(Put_dbg_flag, "doing local put since pkg_len %ld is <= locmemsize %ld\n", pkg_len, locmemsize);
        local_address = (MPI_Aint) DMEM_amalloc((size_t) pkg_len);
        if (local_address == (MPI_Aint) NULL) {
            printf("malloc failed in local put, rank %d, pkg_len = %ld, locmemsize = %ld\n", dmem_rank,
                   pkg_len, locmemsize);
            DMEM_Abort(457);
        }
        /* Check whether new memory overlaps a chunk in locmemtab */
        /* DMEM_checkoverlap(local_address, pkg_len, __LINE__); */

        memcpy((void *) local_address, (void *) pkg_addr, (size_t) pkg_len);
        /* Check that copy was correct
        rc = DMEM_checkcopy(local_address, (MPI_Aint) pkg_addr, pkg_len);
        if (!rc) {
            DMEM_p(1, "checkcopy failed after memcpy in DMEM_Put, rc = %d\n", rc);
            DMEM_Abort(492);
        }
        */
        /* DMEM_Store_address(local_address, pkg_len); replaced by following exchange of messages with own server */
        addrandsize[0] = local_address;
        addrandsize[1] = pkg_len;
        DMEM_p(Put_dbg_flag, "In local put, sending address to server thread\n");
        MPI_Send(addrandsize, 2, MPI_AINT, dmem_rank, DMEM_STORE_ADDR, dmem_comm);
        DMEM_p(Put_dbg_flag, "In local put, sent address to server thread, receiving reply\n");
        MPI_Recv(&dummy1, 1, MPI_INT, dmem_rank, DMEM_STORE_ADDR_RPLY, dmem_comm, &status);
        DMEM_p(Put_dbg_flag, "In local put, received reply from server thread\n");
        /* Create handle for user */
        dh[SRVR] = (MPI_Aint) dmem_rank;
        dh[ADDR] = local_address;
        dh[SIZE] = (MPI_Aint) pkg_len;
        num_local_puts++;
        bytes_local_put += pkg_len;
        /* have own server update locmemsize */
        MPI_Send(&pkg_len, 1, MPI_AINT, dmem_rank, DMEM_DECR_LOCMEM, dmem_comm);
    } else {                      /* non-local, remote put */
        MPI_Status status1;
        int dbg_flag, remote_answer, loop_count;

        dbg_flag = Put_dbg_flag || Put_remote_dbg_flag;
        DMEM_p(dbg_flag, "doing remote put since locmemsize = %ld, pkg_len = %ld\n", locmemsize, pkg_len);
        /* Ask manager where to put it, but first update manager's copy of locmemsize */ 
        /* optimization opportunity: combine the following two exchanges */
        if (dmem_rank != dmem_mgr) { /* don't do this if we ARE the manager process, with memsize 0 */
            MPI_Send(&locmemsize, 1, MPI_AINT, dmem_mgr, DMEM_MEM_UPDATE2, dmem_comm); /* use sep tag since app */
            MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_MUR2, dapp_comm, &status1);
        }
        remote_answer = 0;
        loop_count = 0;
        while (!remote_answer) {
            MPI_Send(&pkg_len, 1, MPI_AINT, dmem_mgr, DMEM_PUT_WHERE_REQ, dmem_comm);
            MPI_Recv(&srvr, 1, MPI_INT, dmem_mgr, DMEM_PUT_WHERE_RPLY, dapp_comm, &status);
            DMEM_p(dbg_flag, "in DMEM_Put, manager suggests %d as target for remote put, rejections = %d\n", srvr, loop_count);
            if (srvr == dmem_rank) { /* if manager suggests this server, we are out of space */
                DMEM_p(1,"in remote put, manager suggested putter (this process) as target - aborting\n");
                DMEM_Abort(392);
            } else {
                /* Ask remote server if it really does have space */
                MPI_Send(&pkg_len, 1, MPI_AINT, srvr, DMEM_PUT, dmem_comm);
                MPI_Recv(&remote_answer, 1, MPI_INT, srvr, DMEM_SPACE_QUERY_RPLY, dapp_comm, &status1);
                if (remote_answer == 1) {
                    MPI_Send(pkg_addr, pkg_len, MPI_BYTE, srvr, DMEM_PUT_DATA, dmem_comm);
                    MPI_Recv(&remote_address, 1, MPI_AINT, srvr, DMEM_PUT_ADDR, dapp_comm, &status);
                    /* Create handle for user */
                    dh[SRVR] = (MPI_Aint) srvr;
                    dh[ADDR] = remote_address;
                    dh[SIZE] = (MPI_Aint) pkg_len;
                } else {
                    loop_count++;
                }
            }
        }
    }

    DMEM_p(Put_dbg_flag, "exiting DMEM_Put, handle = %s, locmemsize = %ld\n", DMEM_Sprint_handle(dh), locmemsize);
    num_puts++;
    bytes_put += pkg_len;
    return 0;
}

/* DMEM_Copy creates a copy of the pkg specified in the first argument
   and returns its handle in the second argument */
int DMEM_Copy(dmem_handle orig, dmem_handle copy)
{
    MPI_Status status;
    MPI_Aint address;
    MPI_Aint addrandsrvr[2];


    DMEM_p(Copy_dbg_flag, "entering copy, handle = %s\n", DMEM_Sprint_handle(orig));
    address = orig[ADDR];
    MPI_Send(&address, 1, MPI_AINT, orig[SRVR], DMEM_COPY, dmem_comm);
    MPI_Recv(addrandsrvr, 2, MPI_AINT, orig[SRVR], DMEM_COPY_RPLY, dapp_comm, &status);

    /* Create handle of copy.  Note server may be different from that of original. */
    copy[ADDR] = addrandsrvr[0];
    copy[SRVR] = addrandsrvr[1];
    copy[SIZE] = orig[SIZE];

    num_copies++;
    bytes_copied += orig[SIZE];
    if (orig[SRVR] == copy[SRVR]) {
        DMEM_p(Copy_dbg_flag, "local copy of %ld bytes\n", orig[SIZE]);
        num_local_copies++;
        bytes_local_copied += orig[SIZE];
    }
    DMEM_p(Copy_dbg_flag, "exiting DMEM_Copy, copy handle = %s\n", DMEM_Sprint_handle(copy));
    return 0;
}

/* DMEM_Get retrieves the package specified by its handle) into
   buf_addr.  Returns -1 if the address in the handle is not found on
   the server in the handle.  The size is in the handle. If we want to
   check for buffer overflow, then we need a buffer size to be added to
   the argument list */
int DMEM_Get(dmem_handle dh, void *buf_addr)
{
    MPI_Status status;
    MPI_Aint addrandsize[2];

    DMEM_p(Get_dbg_flag, "entering DMEM_Get, handle: %s\n", DMEM_Sprint_handle(dh));

    addrandsize[0] = dh[ADDR];
    addrandsize[1] = dh[SIZE];
    MPI_Send(addrandsize, 2, MPI_AINT, dh[SRVR], DMEM_GET, dmem_comm);

    /* poke the buffer to make sure it is OK */
    ((char *) buf_addr)[0] = 'b';
    ((char *) buf_addr)[dh[SIZE] - 1] = 'e';

    MPI_Recv(buf_addr, dh[SIZE], MPI_BYTE, dh[SRVR], MPI_ANY_TAG, dapp_comm, &status);
    if (status.MPI_TAG == DMEM_GET_RPLY) {
        num_gets++;
        bytes_got += dh[SIZE];
        if (dh[SRVR] == dmem_rank)
            bytes_local_got += dh[SIZE];
        DMEM_p(Get_dbg_flag, "exiting DMEM_Get with rc = 0\n");
        return(0);
    }
    else if (status.MPI_TAG == DMEM_GET_RPLY_ERROR) {
        printf("%d got error on get\n", dmem_rank);
        DMEM_p(Get_dbg_flag, "exiting DMEM_Get with rc = -1\n");
        return(-1);
    }
    else {
        printf("unexpected tag %s in message returned to DMEM_Get\n", DMEM_Tag_string(status.MPI_TAG));
        DMEM_Abort(762);
    }
    return(0);
}

/* DMEM_Get_part retrieves a subblock. No check for buffer overflow */
int DMEM_Get_part(dmem_handle dh, MPI_Aint offset, MPI_Aint len, void *buf_addr)
{
    MPI_Status status;
    MPI_Aint handleoffsetlen[DMEM_HANDLE_SIZE+2];

    /* DMEM_checkoverlap((MPI_Aint) buf_addr, len, __LINE__);  check for overlap */

    if (offset + len > dh[SIZE]) {
        printf("Bad arguments passed to DMEM_Get_part by process %d\n", dmem_rank);
        printf("addr = %ld, size = %ld, offset = %ld, len = %ld\n", dh[ADDR], dh[SIZE], offset, len);
        DMEM_Abort(397);
    }
    else {
        handleoffsetlen[0] = dh[0];
        handleoffsetlen[1] = dh[1];
        handleoffsetlen[2] = dh[2];
        handleoffsetlen[3] = offset;
        handleoffsetlen[4] = len;
        MPI_Send(handleoffsetlen, 5, MPI_AINT, dh[SRVR], DMEM_GET_PART, dmem_comm);
        MPI_Recv(buf_addr, len, MPI_BYTE, dh[SRVR], DMEM_GET_PART_DATA, dapp_comm, &status);
    }
    num_get_parts++;
    bytes_got_part += len;
    return 0;
}

/* DMEM_Update replaces a subblock of the pkg with a new subarray */
int DMEM_Update(dmem_handle dh, MPI_Aint offset, MPI_Aint len, void *buf_addr)
{
    MPI_Aint handleoffsetlen[DMEM_HANDLE_SIZE+2];

    if (offset + len > dh[SIZE]) {
        printf("Bad arguments passed to DMEM_Update by process %d\n", dmem_rank);
        printf("addr = %ld, size = %ld, offset = %ld, len = %ld\n", dh[ADDR], dh[SIZE], offset, len);
        DMEM_Abort(395);
    }
    else {
        handleoffsetlen[0] = dh[0];
        handleoffsetlen[1] = dh[1];
        handleoffsetlen[2] = dh[2];
        handleoffsetlen[3] = offset;
        handleoffsetlen[4] = len;
        MPI_Send(handleoffsetlen, 5, MPI_AINT, dh[SRVR], DMEM_UPDATE, dmem_comm);
        MPI_Send(buf_addr, len, MPI_BYTE, dh[SRVR], DMEM_UPDATE_DATA, dmem_comm);
    }
    num_updates++;
    bytes_updated += len;
    return 0;
}

int DMEM_Free(dmem_handle dh)
{
    MPI_Aint addrandsize[2];
    MPI_Status status1;
    addrandsize[0] = dh[ADDR];
    addrandsize[1] = dh[SIZE];
    MPI_Aint dummy;

/*pie     DMEM_p(Free_dbg_flag, "entering DMEM_Free\n"); */
    if (dh[SRVR] == dmem_mgr) {
        printf("Bad handle passed to DMEM_Free (has manager instead of server) by rank %d \n", dmem_rank);
        DMEM_Abort(400);
    }
    /*pie     DMEM_p(Free_dbg_flag, "In DMEM_Free, sending free size %ld to server rank %ld\n", dh[SIZE], dh[SRVR]); */
    MPI_Send(addrandsize, 2, MPI_AINT, dh[SRVR], DMEM_FREE, dmem_comm);
    num_frees++;
    bytes_freed += addrandsize[1];
    if (dh[SRVR] == dmem_rank) {         /* If this is a local free */
        /*pie    DMEM_p(Free_dbg_flag, "local free\n");  */
    }
    /* Wait for confirmation */
    MPI_Recv(&dummy, 1, MPI_AINT, dh[SRVR], DMEM_FREE_CONF, dapp_comm, &status1);
    /*pie    DMEM_p(Free_dbg_flag, "exiting DMEM_Free\n"); */
    return 0;
}

int DMEM_Abort(int abort_code)
{
    DMEM_p(1, "rank %d called DMEM_Abort with code %d\n", dmem_rank, abort_code);
    if (dmem_rank == dmem_mgr)
        DMEM_Dump_gmem_table(globmemtab);
    MPI_Abort(dmem_comm, abort_code);
    return 0;
}

/*-------- Utilities -------------------------------*/

/* This routine should only be called by the manager. */
/* Since there is a line of output for each server, output goes to stderr */
void DMEM_Dump_gmem_table(globmemtabent *globmemtab) /* dump global mem table for debugging */
{
    int i;

    if (dmem_rank == dmem_mgr) {
        fprintf(stderr, " DMEM Global memory table, all processes (approximate):\n");
        for (i = 0; i < dmem_size; i++) {
            fprintf(stderr, "  server rank %d: ", i);
            fprintf(stderr, "memory available = %s; ", n2s(globmemtab[i].size));
            fprintf(stderr, "num_updates = %ld; ", globmemtab[i].num_updates);
            fprintf(stderr, "initial_size = %s\n", n2s(globmemtab[i].initial_size));
        }
    }
    else
        fprintf(stderr, "DMEM_Dump_gmem_table can only be called by the manager, not rank %d \n", dmem_rank);
}
/* This routine should only be called by the manager. */
void DMEM_Dump_gmem_table_stats(globmemtabent *globmemtab) /* print summary data from global mem table */
{
    MPI_Aint total_size, max_size, min_size;
    MPI_Aint total_lwm, max_lwm, min_lwm;
    int total_allocations, max_allocations, min_allocations;
    int total_frees, max_frees, min_frees;
    int i;

    if (dmem_rank != dmem_mgr) {
        DMEM_p(1, "DMEM_Dump_gmem_table_stats should only be called by the manager; aborting\n");
        DMEM_Abort(200);
    }
    total_size = total_lwm = total_allocations = total_frees = 0;
    max_size   = max_lwm =   max_allocations   = max_frees   = 0;
    min_size   = min_lwm =   min_allocations   = min_frees   = (1000 * 1000 * 1000);

    DMEM_Dump_gmem_table(globmemtab);

    for (i = 0; i < dmem_size; i++) {
        if (globmemtab[i].initial_size != 0) {             /* skip processes with 0 memory allocated */
            total_size += globmemtab[i].size;
            if (globmemtab[i].size > max_size)
                max_size = globmemtab[i].size;
            if (globmemtab[i].size < min_size)
                min_size = globmemtab[i].size;
        }
    }

    printf("Manager's view of global memory from globmemtab (approximate, and not including those with 0 memory allocated at Init):\n");
    printf("  Available memory:  ");
    printf("Total = %s, ", n2s(total_size));
    printf("Max = %s, ",   n2s(max_size));
    printf("Min = %s, ",   n2s(min_size));
    printf("Avg (over %d) = %s\n", num_fat_servers, n2s(total_size/num_fat_servers));

}                                                          /* end DMEM_Dump_gmem_table_stats */

char * DMEM_Tag_string(int tag) /* convert tag to string for printing */
{
    switch(tag) {
    case DMEM_GET: return("GET");
    case DMEM_PUT: return("PUT");
    case DMEM_FREE: return("FREE");
    case DMEM_UPDATE: return("UPDATE");
    case DMEM_SHUTDOWN: return("SHUTDOWN");
    case DMEM_PUT_WHERE_REQ: return("PUT_WHERE_REQUEST");
    case DMEM_PUT_WHERE_RPLY: return("PUT_WHERE_REPLY");
    case DMEM_GET_RPLY: return("GET_REPLY");
    case DMEM_GET_RPLY_ERROR: return("GET_REPLY_ERROR");
    case DMEM_PUT_ADDR: return("PUT_ADDR");
    case DMEM_INIT_SIZE: return("INIT_SIZE");
    case DMEM_DUMP_GLOBMEMTAB: return("DUMP_GLOBMEMTAB");
    case DMEM_UPDATE_RPLY: return("UPDATE_REPLY");
    case DMEM_UPDATE_DATA: return("UPDATE_DATA");
    case DMEM_COPY: return("COPY");
    case DMEM_COPY_RPLY: return("COPY_RPLY");
    case DMEM_GET_PART: return("GET_PART");
    case DMEM_GET_PART_DATA: return("GET_PART_DATA");
    case DMEM_ACK: return("ACK");
    case DMEM_MEM_QUERY: return("MEM_QUERY");
    case DMEM_MEM_QUERY_REPLY: return("MEM_QUERY_REPLY");
    case DMEM_MEM_FREE: return("MEM_FREE");
    case DMEM_STATS: return("STATS");
    case DMEM_FREE_CONF: return("FREE_CONF");
    case DMEM_MEM_UPDATE: return("MEM_UPDATE");
    case DMEM_MEM_UPDATE2: return("MEM_UPDATE2");
    case DMEM_MUR: return("MEM_UPDATE_REPLY");
    case DMEM_MUR2: return("MEM_UPDATE_REPLY2");
    case DMEM_DECR_LOCMEM: return("DECR_LOCMEM");
    case DMEM_COPY_WHERE_REQ: return("DMEM_COPY_WHERE_REQ");
    case DMEM_COPY_WHERE_RPLY: return("DMEM_COPY_WHERE_RPLY");
    case DMEM_COPY_DATA: return("COPY_DATA");
    case DMEM_SPACE_QUERY_RPLY: return("SPACE_QUERY_RPLY");
    case DMEM_PUT_DATA: return("PUT_DATA");
    case DMEM_COPY_QUERY: return("COPY_QUERY");
    case DMEM_MIN_LWM: return("MIN_LWM");
    case DMEM_STORE_ADDR: return("STORE_ADDR");
    case DMEM_STORE_ADDR_RPLY: return("STORE_ADDR_RPLY");
    default:
        sprintf(unknown_tag, "%d", tag);
        return(unknown_tag);
    }
}

/* find which server has the most memory available */
int DMEM_Find_max(globmemtabent *globmemtab, int excluded)
{
    int i, maxloc;
    MPI_Aint max_so_far;

    maxloc = -1;
    max_so_far = 0;
    for (i = 0; i < dmem_size; i++) {
        if ((i != dmem_mgr) && (i != excluded)) {          /* skip manager and excluded*/
            if (globmemtab[i].size > max_so_far) {   
                maxloc = i;
                max_so_far = globmemtab[i].size;
            }
        }
    }
    return(maxloc);
}

/* find a slot in the server's local memory table and store an address and size */
void DMEM_Store_address(MPI_Aint addr, MPI_Aint size)
{
    int i, flag = 0;

    for (i = 0; i < locmemtabsize; i++) {
        /* change 0 in next line to appropriate debug flag, depending on problem being investigated */
        DMEM_p(0, "In DMEM_Store_address, rank %d, i = %d, locmemtab[i].address = %ld, usage_flag = %d\n",
               dmem_rank, i, locmemtab[i].address, locmemtab[i].usage_flag);  
        if ((locmemtab[i].usage_flag == -1) || (locmemtab[i].usage_flag == -2))
            break;
    }
    if (i == locmemtabsize) {
        fprintf(stderr, "DMEM: local memory table full on server rank %d, table size = %d\n", dmem_rank, locmemtabsize);
        /* DMEM_Dumplocmemtab(); */
        /* DMEM_Abort(345); */
        fprintf(stderr, "rank %d expanding locmemtab\n", dmem_rank);
        DMEM_Expandlocmemtab();
        flag  = 1;
    }

    locmemtab[i].address    = addr;
    locmemtab[i].size       = size;
    locmemtab[i].usage_flag = -3;
    if (i > locmemtabsize_hwm)
        locmemtabsize_hwm++;
    DMEM_p(0, "locmemtabsize_hwm incremented to %d\n", locmemtabsize_hwm);
    /*
    if (flag) {
        printf("rank %d after expanding locmemtab\n", dmem_rank);
        DMEM_Dumplocmemtab();
    }
    */
}

/* send message to manager to have it dump the global memory table.
   Wait for ack to keep output together. */
int DMEM_Dumpglobmemtab()
{
    MPI_Aint   dummy;
    MPI_Status status;

    MPI_Send(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_DUMP_GLOBMEMTAB, dmem_comm);
    MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_ACK, dapp_comm, &status); /* receive ack */
    return 0;
}

/* Print statistics from global memory table */
int DMEM_Stats_snapshot()
{
    MPI_Aint dummy;
    MPI_Status status;

    MPI_Send(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_STATS, dmem_comm);
    MPI_Recv(&dummy, 1, MPI_AINT, dmem_mgr, DMEM_ACK, dapp_comm, &status); /* receive ack */
    return 0;
}

int DMEM_Dumplocmemtab()
{
    int i;

    fprintf(stderr, "DMEM: Local memory table for server rank %d: init (max) = %ld; lwm = %ld, currently available = %ld\n",
            dmem_rank, locmemsize_hwm, locmemsize_lwm, locmemsize);
    for (i = 0; i < locmemtabsize; i++) {
        if ((locmemtab[i].usage_flag == -2) || (locmemtab[i].usage_flag == -3))
            fprintf(stderr, "DMEM srvr = %d, entry = %d, addr = %s, size = %s, flag = %d\n",
                    dmem_rank, i, a2s(locmemtab[i].address), a2s(locmemtab[i].size),locmemtab[i].usage_flag);
    }
    return 0;
}

/* find an entry in the local memory table and return the index */
int DMEM_Lookup(MPI_Aint address)
{
    int i;

    for (i = 0; i < locmemtabsize; i++)
        if ((locmemtab[i].address == address) && (locmemtab[i].usage_flag == -3)) /* valid addr and in use */
            return(i);
    return(-1);
}

int DMEM_Print_handle(dmem_handle dh)
{
    fprintf(stderr, "srvr rank %ld , addr=%ld, size=%ld",
            dh[SRVR], dh[ADDR], dh[SIZE]);
    return 0;
}

char * DMEM_Sprint_handle(dmem_handle dh)
{
    static char handle_string[100];

    sprintf(handle_string, "srvr rank %ld , addr=%ld, size=%ld",
            dh[SRVR], dh[ADDR], dh[SIZE]);
    return(handle_string);
}

/* handle access routines */
int DMEM_Get_server(dmem_handle dh)
{
    return (int) dh[SRVR];
}

MPI_Aint DMEM_Get_address(dmem_handle dh)
{
    return dh[ADDR];
}

MPI_Aint DMEM_Get_size(dmem_handle dh)
{
    return dh[SIZE];
}

int DMEM_Stats_final()
{
    printf("DMEM_Stats_final is obsolete, merged into DMEM_Finalize\n");
    return 0;
}

/* formatting routines for addresses and long integers */

/* address to string - insert commas into long integers (like MPI_Aints) */
/* also used to insert commas into counts */
char* a2s(long n)
{
    char buf[30];
    int i, j, ctr;
    char *retbuf;

    /* printf("entering a2s, n = %ld\n", n); */
    retbuf = (char *) DMEM_amalloc(30);
    sprintf(buf, "%ld", n);
    i = strlen(buf);            /* buf[i] = '\0' */
    j = 29;
    retbuf[j] = '\0';
    j--;
    ctr = 0;

    for (i = strlen(buf)-1; i >=0; i--) {
        retbuf[j] = buf[i];
        ctr++;
        if ((ctr % 3 == 0) && (i != 0) && (buf[i-1] != '-')) {
            j--;
            retbuf[j] = ',';
        }
        j--;
    }
    /* printf("exiting a2s, return value = %s\n", &retbuf[j+1]); */
    return(&retbuf[j+1]);
    /*    return(&retbuf[++j]); */
}

/* memory size to string - round to nearest megabyte if bigger than 2M */
char* n2s(long n)
{
#define MEGA 1024*1024

    long megabytes;
    char *result;
    result = (char * ) DMEM_amalloc(30);
    if (n > 2*MEGA) {
        megabytes = n / (MEGA); /* parentheses necessary */
        if ((n % (MEGA)) > (.5 * (MEGA)))
            megabytes++;
        sprintf(result, "%s MB", a2s(megabytes)); /* don't use n2s -- recursive */
    } else
        sprintf(result, "%s", a2s(n)); /* likewise */
    return(result);
}

/* Note this doesn't put the final \n on; you have to include it in the message */
void DMEM_p_fcn(int flag, char *fmt, ...)
{
    char *s;
    va_list ap;

    if ( !dmem_dbgprintf_flag)      /* global */
        return;
    if ( !flag)                     /* local  */
        return;

    va_start(ap, fmt);
    vasprintf(&s, fmt, ap);
    va_end(ap);

    if ( !s ) {
        fprintf(stderr, "DMEM: **** FAILED TO ALLOCATE MEMORY FOR DEBUG PRINT ****\n");
        fflush(stderr);
        return;
    }
    fprintf(stderr,"%f: DMEM rank %d %s", MPI_Wtime(), dmem_rank, s);
    fflush(stderr);
    free(s);
}

/* Check that a local copy (for either local put or local copy was done correctly by memcpy */
/* Returns 1 if the copies are identical, 0 if not */
int DMEM_checkcopy(MPI_Aint buf1, MPI_Aint buf2, MPI_Aint len)
{
    MPI_Aint i;

    DMEM_p(Memory_dbg_flag, "entering checkcopy, buf1 = %ld, buf2 = %ld, len = %ld\n", buf1, buf2, len);
    if ((memcmp((void *) buf1, (void *) buf2, (size_t) len)) != 0) {
        for (i = 0; i < len; i++) {
            if (*((char *) (buf1 + i)) != *((char *) (buf2 + i))) {
                DMEM_p(1, "checkcopy failed at i = %ld\n", i);
                return 0;
            }
        }
    }
    DMEM_p(Memory_dbg_flag, "leaving checkcopy, len = %ld\n", len);
    return 1;
}

/* Check whether a new allocation overlaps with existing ones in use in locmemtab returns 1 if OK, 0 if not*/
/* relies on __LINE__ for value of linenum argument */

int DMEM_checkoverlap(MPI_Aint addr, MPI_Aint size, int linenum)
{
    int i, result;
    result = 1;
    for (i = 0; i < locmemtabsize_hwm; i++) {
        if (locmemtab[i].usage_flag == -3) { /* entry in use */
            if (DMEM_allocchk(locmemtab[i].address, locmemtab[i].size, addr, size)) {
                result = 0;
                DMEM_p(1, "server found memory overlap for (%ld,%ld) with existing (%ld,%ld) at index %d, line %d\n",
                       addr, size, locmemtab[i].address, locmemtab[i].size, i, linenum);
                DMEM_Abort(200);
            }
        }
    }
    return(result);
}
                    

/* Check whether two buffers overlap, returns 0 if they don't, 1 if they do */
int DMEM_allocchk(MPI_Aint addr1, MPI_Aint size1, MPI_Aint addr2, MPI_Aint size2)
{
    int result;

    if (((addr1 + size1) < addr2) || (addr1 > (addr2 + size2)))
        result = 0;
    else
        result = 1;
    /*printf("testing (%ld, %ld) and (%ld, %ld): result is %d\n", addr1, size1, addr2, size2, result); */
    return(result);
}

/* align all mallocs, to make them better for BG/Q, copied from BG/Q wiki */
void *DMEM_amalloc(size_t n)
{
    int rc;
    void * ptr;
    size_t alignment = 32; /* 128 might be better since that ensures every heap allocation 
                            * starts on a cache-line boundary */
    rc = posix_memalign( &ptr , alignment , n );
    
    if (rc == ENOMEM) {
        printf("ENOMEM failure in posix_memalign, size = %ld\n", (MPI_Aint) n);
        return NULL;
    } else if (rc == EINVAL) {
        printf("EINVAL failure in posix_memalign, size = %ld, alignment = %ld\n", (MPI_Aint) n, (MPI_Aint) alignment);
        return NULL;
    } else
        return ptr;
}

/* double the size of locmemtab by allocating a new table and copying the old contents into it */
int DMEM_Expandlocmemtab()
{
    int i, newsize;
    locmemtabent *newlocmemtab;

    fprintf(stderr, "DMEM: rank %d entering DMEM_Expand_locmemtab, locmemtabsize = %d\n", dmem_rank, locmemtabsize);

    newsize = 2 * locmemtabsize;
    newlocmemtab = DMEM_amalloc(newsize * sizeof(locmemtabent));
    for (i = 0; i < locmemtabsize; i++) {
        newlocmemtab[i].address    = locmemtab[i].address;
        newlocmemtab[i].size       = locmemtab[i].size;
        newlocmemtab[i].usage_flag = locmemtab[i].usage_flag;
    }
    for (i = locmemtabsize; i < newsize; i++) {
        newlocmemtab[i].address    = -1;
        newlocmemtab[i].size       = -1;
        newlocmemtab[i].usage_flag = -1;
    }
    locmemtabsize = newsize;
    free(locmemtab);            /* free old table */
    locmemtab = newlocmemtab;

    fprintf(stderr, "DMEM rank %d leaving DMEM_Expand_locmemtab, locmemtabsize = %d\n", dmem_rank, locmemtabsize);
    return 0;
}


/* Initialize debug print and other flags, which can also be modified during the run. */
/* Setting a flag to 1 will turn on tracing for a family of events
   associated with a particular situation, on both the applcation and
   server side.  The basic ones are for the basic API functions.  The
   manager and server flags trace all calls to manager or server.
   This routine is called after dmem_rank has been set, so these flag assignments
   can be guarded as in "if (dmem_rank == 241) {....}" if desired.
   All tracing can be compiled out by setting DMEM_TRACE to 0 near top of this file. 
*/
void DMEM_Init_dbg_flags()
{
    /* For the dbg flags, 1 means print related trace messages on stderr, 0 means don't. */
    Get_dbg_flag         = 0;
    Put_dbg_flag         = 0;
    Put_remote_dbg_flag  = 0;
    Copy_dbg_flag        = 0;
    Copy_remote_dbg_flag = 0;
    Free_dbg_flag        = 0;
    Get_part_dbg_flag    = 0;
    Update_dbg_flag      = 0;
    Server_dbg_flag      = 0;
    Manager_dbg_flag     = 0;
    Memory_dbg_flag      = 0;
    Error_return_flag    = 0;           /* if a bad address is detected, 0 -> abort; 1 -> return -1 */
    Update_counter_init  = 5;           /* how often a server updates the manager on its locmemsize */
    Selection_algorithm  = ORIGINAL;    /* How selection of a server to store a put or copy is done.  Values are:
                                          *    ORIGINAL - local put or copy if possible, otherwise ask manager,
                                         *         who recommends server with most memory available
                                         *    NOLOCAL  - no local copies    (local puts allowed)
                                         * THIS IS NOT YET IMPLEMENTED; only ORIGINAL.
                                         */
}

int DMEM_Set_errors_return(int value)
{
    Error_return_flag = value;
    return 0;
}
